@@ -1,41 +1,30 @@
 // Take a look at the license at the top of the repository in the LICENSE file.
 
-// rustdoc-stripper-ignore-next
-//! `Task` bindings.
-//!
-//! The API distinguishes between `LocalTask` and `Task`, where the latter can be used
-//! with threads, but has more strict requirements, in particular it requires its value,
-//! source object, and callbacks must implement the `Send` trait.
-//!
-//! The `Task` and `LocalTask` types provide idiomatic access to glib's `GTask` API, for
-//! instance by being generic over their value type, while not completely departing
-//! from the underlying C API.
-//! Unfortunately this API does not allow to automatically enforce all the
-//! invariants required to be a completely safe abstraction.
-//!
-//! For this reason, the constructors of `Task` and `LocalTask` are marked as
-//! unsafe and when using these types the caller is responsible to ensure the
-//! following requirements are satisfied
-//!
-//! * You should not create a `LocalTask`, upcast it to a `glib::Object` and then
-//!   downcast it to a `Task`, as this will bypass the thread safety requirements
-//! * You should ensure that the `return_result`, `return_error_if_cancelled` and
-//!   `propagate()` methods are only called once.
+use std::{boxed::Box as Box_, future::Future, mem::transmute, panic, ptr};
 
-use crate::AsyncResult;
-use crate::Cancellable;
-use glib::object::IsA;
-use glib::object::ObjectType as ObjectType_;
-use glib::signal::connect_raw;
-use glib::signal::SignalHandlerId;
-use glib::translate::*;
-use glib::value::ValueType;
-use glib::Cast;
-use std::boxed::Box as Box_;
-use std::mem::transmute;
-use std::ptr;
+use glib::{
+    prelude::*,
+    signal::{connect_raw, SignalHandlerId},
+    translate::*,
+    value::ValueType,
+};
+
+use futures_channel::oneshot;
+
+use crate::{AsyncResult, Cancellable};
 
 glib::wrapper! {
+    // rustdoc-stripper-ignore-next
+    /// `LocalTask` provides idiomatic access to gio's `GTask` API, for
+    /// instance by being generic over their value type, while not completely departing
+    /// from the underlying C API. `LocalTask` does not require its value to be `Send`
+    /// and `Sync` and thus is useful to to implement gio style asynchronous
+    /// tasks that run in the glib main loop. If you need to run tasks in threads
+    /// see the `Task` type.
+    ///
+    /// The constructors of `LocalTask` and `Task` is marked as unsafe because this API does
+    /// not allow to automatically enforce all the invariants required to be a completely
+    /// safe abstraction. See the `Task` type for more details.
     #[doc(alias = "GTask")]
     pub struct LocalTask<V: ValueType>(Object<ffi::GTask, ffi::GTaskClass>) @implements AsyncResult;
 
@@ -45,6 +34,23 @@ glib::wrapper! {
 }
 
 glib::wrapper! {
+    // rustdoc-stripper-ignore-next
+    /// `Task` provides idiomatic access to gio's `GTask` API, for
+    /// instance by being generic over their value type, while not completely departing
+    /// from the underlying C API. `Task` is `Send` and `Sync` and requires its value to
+    /// also be `Send` and `Sync`, thus is useful to to implement gio style asynchronous
+    /// tasks that run in threads. If you need to only run tasks in glib main loop
+    /// see the `LocalTask` type.
+    ///
+    /// The constructors of `LocalTask` and `Task` is marked as unsafe because this API does
+    /// not allow to automatically enforce all the invariants required to be a completely
+    /// safe abstraction. The caller is responsible to ensure the following requirements
+    /// are satisfied
+    ///
+    /// * You should not create a `LocalTask`, upcast it to a `glib::Object` and then
+    ///   downcast it to a `Task`, as this will bypass the thread safety requirements
+    /// * You should ensure that the `return_result`, `return_error_if_cancelled` and
+    ///   `propagate()` methods are only called once.
     #[doc(alias = "GTask")]
     pub struct Task<V: ValueType + Send>(Object<ffi::GTask, ffi::GTaskClass>) @implements AsyncResult;
 
@@ -55,7 +61,7 @@ glib::wrapper! {
 
 macro_rules! task_impl {
     ($name:ident $(, @bound: $bound:tt)? $(, @safety: $safety:tt)?) => {
-        impl <V: ValueType $(+ $bound)?> $name<V> {
+        impl <V: Into<glib::Value> + ValueType $(+ $bound)?> $name<V> {
             #[doc(alias = "g_task_new")]
             #[allow(unused_unsafe)]
             pub unsafe fn new<S, P, Q>(
@@ -101,7 +107,7 @@ macro_rules! task_impl {
 
             #[doc(alias = "g_task_get_cancellable")]
             #[doc(alias = "get_cancellable")]
-            pub fn cancellable(&self) -> Cancellable {
+            pub fn cancellable(&self) -> Option<Cancellable> {
                 unsafe { from_glib_none(ffi::g_task_get_cancellable(self.to_glib_none().0)) }
             }
 
@@ -118,8 +124,8 @@ macro_rules! task_impl {
                 }
             }
 
-            #[cfg(any(feature = "v2_60", feature = "dox"))]
-            #[cfg_attr(feature = "dox", doc(cfg(feature = "v2_60")))]
+            #[cfg(feature = "v2_60")]
+            #[cfg_attr(docsrs, doc(cfg(feature = "v2_60")))]
             #[doc(alias = "g_task_set_name")]
             pub fn set_name(&self, name: Option<&str>) {
                 unsafe {
@@ -175,8 +181,8 @@ macro_rules! task_impl {
                 unsafe { from_glib_none(ffi::g_task_get_context(self.to_glib_none().0)) }
             }
 
-            #[cfg(any(feature = "v2_60", feature = "dox"))]
-            #[cfg_attr(feature = "dox", doc(cfg(feature = "v2_60")))]
+            #[cfg(feature = "v2_60")]
+            #[cfg_attr(docsrs, doc(cfg(feature = "v2_60")))]
             #[doc(alias = "g_task_get_name")]
             #[doc(alias = "get_name")]
             pub fn name(&self) -> Option<glib::GString> {
@@ -249,19 +255,20 @@ macro_rules! task_impl {
                     Ok(v) => unsafe {
                         ffi::g_task_return_value(
                             self.to_glib_none().0,
-                            v.to_value().to_glib_full() as *mut _,
+                            v.to_value().to_glib_none().0 as *mut _,
                         )
                     },
                     #[cfg(not(feature = "v2_64"))]
                     Ok(v) => unsafe {
+                        let v: glib::Value = v.into();
                         ffi::g_task_return_pointer(
                             self.to_glib_none().0,
-                            v.to_value().to_glib_full() as *mut _,
+                            <glib::Value as glib::translate::IntoGlibPtr::<*mut glib::gobject_ffi::GValue>>::into_glib_ptr(v) as glib::ffi::gpointer,
                             Some(value_free),
                         )
                     },
                     Err(e) => unsafe {
-                        ffi::g_task_return_error(self.to_glib_none().0, e.to_glib_full() as *mut _);
+                        ffi::g_task_return_error(self.to_glib_none().0, e.into_glib_ptr());
                     },
                 }
             }
@@ -376,11 +383,74 @@ impl<V: ValueType + Send> Task<V> {
 unsafe impl<V: ValueType + Send> Send for Task<V> {}
 unsafe impl<V: ValueType + Send> Sync for Task<V> {}
 
+// rustdoc-stripper-ignore-next
+/// A handle to a task running on the I/O thread pool.
+///
+/// Like [`std::thread::JoinHandle`] for a blocking I/O task rather than a thread. The return value
+/// from the task can be retrieved by awaiting on this handle. Dropping the handle "detaches" the
+/// task, allowing it to complete but discarding the return value.
+#[derive(Debug)]
+pub struct JoinHandle<T> {
+    rx: oneshot::Receiver<std::thread::Result<T>>,
+}
+
+impl<T> JoinHandle<T> {
+    #[inline]
+    fn new() -> (Self, oneshot::Sender<std::thread::Result<T>>) {
+        let (tx, rx) = oneshot::channel();
+        (Self { rx }, tx)
+    }
+}
+
+impl<T> Future for JoinHandle<T> {
+    type Output = std::thread::Result<T>;
+    #[inline]
+    fn poll(
+        mut self: std::pin::Pin<&mut Self>,
+        cx: &mut std::task::Context<'_>,
+    ) -> std::task::Poll<Self::Output> {
+        std::pin::Pin::new(&mut self.rx)
+            .poll(cx)
+            .map(|r| r.unwrap())
+    }
+}
+
+impl<T> futures_core::FusedFuture for JoinHandle<T> {
+    #[inline]
+    fn is_terminated(&self) -> bool {
+        self.rx.is_terminated()
+    }
+}
+
+// rustdoc-stripper-ignore-next
+/// Runs a blocking I/O task on the I/O thread pool.
+///
+/// Calls `func` on the internal Gio thread pool for blocking I/O operations. The thread pool is
+/// shared with other Gio async I/O operations, and may rate-limit the tasks it receives. Callers
+/// may want to avoid blocking indefinitely by making sure blocking calls eventually time out.
+///
+/// This function should not be used to spawn async tasks. Instead, use
+/// [`glib::MainContext::spawn`] or [`glib::MainContext::spawn_local`] to run a future.
+pub fn spawn_blocking<T, F>(func: F) -> JoinHandle<T>
+where
+    T: Send + 'static,
+    F: FnOnce() -> T + Send + 'static,
+{
+    // use Cancellable::NONE as source obj to fulfill `Send` requirement
+    let task = unsafe { Task::<bool>::new(Cancellable::NONE, Cancellable::NONE, |_, _| {}) };
+    let (join, tx) = JoinHandle::new();
+    task.run_in_thread(move |_, _: Option<&Cancellable>, _| {
+        let res = panic::catch_unwind(panic::AssertUnwindSafe(func));
+        let _ = tx.send(res);
+    });
+
+    join
+}
+
 #[cfg(test)]
 mod test {
     use super::*;
-    use crate::prelude::*;
-    use crate::test_util::run_async_local;
+    use crate::{prelude::*, test_util::run_async_local};
 
     #[test]
     fn test_int_async_result() {
@@ -430,7 +500,7 @@ mod test {
 
         impl MySimpleObject {
             pub fn new() -> Self {
-                glib::Object::new(&[]).expect("Failed to create MySimpleObject")
+                glib::Object::new()
             }
 
             #[doc(alias = "get_size")]
